#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Joseph Helm and Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.BinPacking2D {
  [Item("Swap2MoveEvaluator", "Move evaluator for 2-opt moves.")]
  [StorableClass]
  public sealed class Swap2MoveEvaluator : SingleSuccessorOperator,
    ISingleObjectiveMoveEvaluator, ISingleObjectiveMoveOperator, IPermutationSwap2MoveOperator {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<Permutation> PermutationParameter {
      get { return (ILookupParameter<Permutation>)Parameters["PackingSequence"]; }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<DoubleValue> MoveQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["MoveQuality"]; }
    }
    public ILookupParameter<ReadOnlyItemList<PackingItem>> PackingItemMeasuresParameter {
      get { return (LookupParameter<ReadOnlyItemList<PackingItem>>)Parameters["Items"]; }
    }
    public ILookupParameter<PackingShape> PackingBinMeasuresParameter {
      get { return (LookupParameter<PackingShape>)Parameters["BinShape"]; }
    }
    public ILookupParameter<IPermutationDecoder> PackingSolutionDecoderParameter {
      get { return (ILookupParameter<IPermutationDecoder>)Parameters["Decoder"]; }
    }
    public ILookupParameter<IEvaluator> PackingPlanEvaluatorParameter {
      get { return (ILookupParameter<IEvaluator>)Parameters["SolutionEvaluator"]; }
    }

    public ILookupParameter<Swap2Move> Swap2MoveParameter {
      get { return (ILookupParameter<Swap2Move>)Parameters["Swap2Move"]; }
    }

    [StorableConstructor]
    private Swap2MoveEvaluator(bool deserializing) : base(deserializing) { }
    private Swap2MoveEvaluator(Swap2MoveEvaluator original, Cloner cloner) : base(original, cloner) { }
    public Swap2MoveEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<Permutation>("PackingSequence", "The encoded solution candidate to evaluate"));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality of a packing solution."));
      Parameters.Add(new LookupParameter<DoubleValue>("MoveQuality", "The evaluated quality of a move on a packing solution."));
      Parameters.Add(new LookupParameter<ReadOnlyItemList<PackingItem>>("Items", "Packing-item data taken from the bin-packing problem-instance."));
      Parameters.Add(new LookupParameter<PackingShape>("BinShape", "Packing-bin data taken from the bin-packing problem-instance."));
      Parameters.Add(new LookupParameter<IPermutationDecoder>("Decoder", "The decoding operator that is used to calculate a packing plan from the used representation."));
      Parameters.Add(new LookupParameter<IEvaluator>("SolutionEvaluator", "The actual packing plan evaluation operator."));
      Parameters.Add(new LookupParameter<Swap2Move>("Swap2Move", "The move to evaluate."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Swap2MoveEvaluator(this, cloner);
    }


    public override IOperation Apply() {
      var move = Swap2MoveParameter.ActualValue;
      var permutation = PermutationParameter.ActualValue;
      var binShape = PackingBinMeasuresParameter.ActualValue;
      var items = PackingItemMeasuresParameter.ActualValue;

      // uses full evaluation
      Swap2Manipulator.Apply(permutation, move.Index1, move.Index2);
      var solution = PackingSolutionDecoderParameter.ActualValue.Decode(permutation, binShape, items);

      var quality = PackingPlanEvaluatorParameter.ActualValue.Evaluate(solution);

      double moveQuality = quality;

      if (MoveQualityParameter.ActualValue == null)
        MoveQualityParameter.ActualValue = new DoubleValue(moveQuality);
      else
        MoveQualityParameter.ActualValue.Value = moveQuality;
      return base.Apply();
    }
  }
}
